bits 16

%if 1
%define UNVEIL_ON_GAME_OVER	+16
%define CGA_DISPLAY		+6
%define MDA_SUPPORT		+74
%define DOS_QUIT		+12
%define UNDO_FLAG		+10
%define EXPLOSION		+4
;%define LAZY_CHECK		-4  cylinder board
;%define USE_RDTSC		-15 need a 586+
%else
%define USE_RDTSC		-15 need a 586+
%endif

cpu 8086
org 0x7c00

;; Constants

;; Boot sector size in bytes
%assign BootSector.Size 512

;; Words in 16 bit x86 are 2 bytes
%assign WordSize 2

;; This is the value to store in segment register to access the VGA text buffer.
;; In 16 bit x86, segmented memory accesses are of the form:
;;
;;   (segment register) * 0x10 + (offset register)
;;
;; The VGA text buffer is at 0xb80000, so if 0xb800 is stored in a segment
;; register, then memory access instructions will be relative to the VGA text
;; buffer, allowing easier access. For example, trying to access the nth byte of
;; memory will *actually* access the nth byte of the text buffer.
%ifdef MDA_SUPPORT
%assign TextBuf.Seg 0xb000
%else
%assign TextBuf.Seg 0xb800
%endif

;; Dimensions of text buffer
%assign TextBuf.Width 40
%assign TextBuf.Height 25
%assign TextBuf.Size (TextBuf.Width * TextBuf.Height)

;; Macro to get the index of a text buffer cell from coordinates
%define TextBuf.Index(y, x) ((y) * TextBuf.Width * 2 + (x) * 2)

;; Keyboard scan codes
;; http://www.delorie.com/djgpp/doc/rbinter/it/06/0.html
%assign Key.ScanCode.Space    0x39
%assign Key.ScanCode.Up       0x48
%assign Key.ScanCode.Down     0x50
%assign Key.ScanCode.Left     0x4b
%assign Key.ScanCode.Right    0x4d
%assign Key.ScanCode.Enter    0x1c
%assign Key.ScanCode.QuitGame 0x01

;; This is a convenience macro for creating VGA characters. VGA characters are
;; 16 bit words, with the lower byte as the ASCII value and the upper byte
;; holding the foreground and background colors.
%define VgaChar(color, ascii) (((color) << 8) | (ascii))

;; VGA colors to use for game items
;; https://wiki.osdev.org/Text_UI#Colours
%assign Color.Veiled         0x77
%assign Color.Unveiled       0xf0
%assign Color.Cursor         0x00
%assign Color.Flag           0xcc
%assign Color.GameWinText    0x20
%assign Color.GameOverText   0xc0

%ifdef MDA_SUPPORT
%assign Mda.InverseAttr      0x70 ; 0x[7F][08]	Bit 3: High intensity.
%assign Mda.CursorAttr       0x09 ; 0x?[19]	Bits 0-2: 1 => underline, other values => no underline.
%assign Mda.HiddenAttr       0x00 ; 0x[08][08]	Bit 7: Blink.
%assign Mda.Veiled           0xdb ; 0xb1
%assign Mda.Flag             0x02
%assign Mda.Cursor           0xb1 ; 0x16
%assign Mda.Screen.Offset    0x8000
%assign Mda.Buffer           0x8000
%define Mda.Attr(cga, mda)   (((cga) << 8) | (mda))
%endif

;; This value is used to calculate bomb frequency. The probability that any
;; given cell is a bomb is (1/2)^n, where n = "number of ones in the binary
;; representation of BombFreq".
;;
;; In other words, when BombFreq=-1, every cell is a bomb, and appending a one
;; halves the amount of bombs.
%assign BombFreq             0b111
%assign Ascii.Bomb           '*'

;; BootMine is supported as both a DOS game and a boot sector game :)

;; Entry point: set up graphics and run game
BootMine:
  cld
  sti                   ;allow interrupts

  ; VGA text mode 0x00
  ; 320x200 pixel resolution
  ; 40x25 text resolution
  ; 16 colors
  ; http://www.delorie.com/djgpp/doc/rbinter/id/74/0.html
  xor ax, ax
  int 0x10

%ifdef CGA_DISPLAY
  ; Toggle intensity/blinking bit
  ; http://www.techhelpmanual.com/140-int_10h_1003h__select_foreground_blink_or_bold_background.html
  xor bx, bx		;blinking off
  mov ds, bx
  mov ax, 0x1003	;toggle intensity/blinking bit (Jr, PS, TANDY 1000, EGA, VGA)
  int 0x10
  mov si, 0x463
  lodsw			;get port address for 6845 video controller chip
  add al, 4
  xchg ax, dx
  and byte [si], 0xdf	;mask value by 1101 1111 (to clear bit 5)
  lodsb			;get new value of Mode Select Register
  out dx, al		;disable blink (set for bold background)
%else
  ; Disable blinking text
  ; https://www.reddit.com/r/osdev/comments/70fcig/blinking_text/dn2t6u8?utm_source=share&utm_medium=web2x
  ; Read I/O Address 0x03DA to reset index/data flip-flop
  mov dx, 0x03DA
  in al, dx
  ; Write index 0x30 to 0x03C0 to set register index to 0x30
  mov dl, 0xC0
  mov al, 0x30
  out dx, al
  ; Read from 0x03C1 to get register contents
  inc dx
  in al, dx
  ; Unset Bit 3 to disable Blink
  and al, 0xF7
  ; Write to 0x03C0 to update register with changed value
  dec dx
  out dx, al
%endif

  ; Disable VGA text mode cursor
  ; https://wiki.osdev.org/Text_Mode_Cursor#Disabling_the_Cursor
  mov ah, 0x01
  mov ch, 0x3f
  int 0x10

;; Run game (the game is restarted by jumping here)
RunGame:
  push cs
  pop ss
  mov sp, TextBuf.Seg

  ; Load VGA text buffer segment into segment registers
  mov es, sp
  mov ds, sp

%ifndef USE_RDTSC
  ; Initialyze the simple pseudo-random number generator
  ; seed = set_system_time()
  mov ah, 0
  int 0x1a
%endif

;; Set all cells of game map to veiled '0' cells
ZeroTextBuf:
%ifdef MDA_SUPPORT
 %ifdef UNVEIL_ON_GAME_OVER
  call FillScreen
 %else
  mov ax, VgaChar(Color.Veiled, '0')
  mov di, Mda.Screen.Offset
  mov bx, TextBuf.Width - Mda.Screen.Offset
  mov bp, TextBuf.Height
.1:
  mov cx, TextBuf.Width
.2:
  mov [cs:di + Mda.Buffer - Mda.Screen.Offset], ax
  mov byte [bx + di], Mda.Veiled
  stosw
  loop .2
  add bx, TextBuf.Width * 2
  dec bp
  jnz .1
 %endif
%else
  mov ax, VgaChar(Color.Veiled, '0')
  mov cx, TextBuf.Size
  xor di, di
  rep stosw
%endif

;; Populate text buffer with mines and digits
;;
;; This is done with a single triple-nested loop. The nested loops iterate over
;; y coordinates, then x coordinates, then over the 8 adjacent cells at (y, x).
;;
;; Inside the 2nd loop level is bomb generation logic. Digit incrementing logic
;; is in the 3rd loop level.
;;
;; Note that the coordinates on the outside border are skipped to avoid bounds
;; checking logic.
PopulateTextBuf:
  ; Iterate over y coordinates. ch = 0 form ZeroTextBuf
  mov cl, TextBuf.Height - 2

.LoopY:
  mov di, TextBuf.Width - 2

.LoopX:

  ; The register al holds a boolean that is 1 if the current cell is a bomb, 0
  ; otherwise.
%ifndef USE_RDTSC
  ; It is calculated by bitwise and-ing the result of the simple pseudo-random number
  ; generator seed = ((seed + LARGE_PRIME1) * LARGE_PRIME2) % LARGE_PRIME3
  ;
  ; al = ! (prng() & BombFreq)
%assign Prime1 32749
%assign Prime2 65519
%assign Prime3 65521
  xchg ax, dx
  add ax, Prime1
  mov si, Prime2
  mul si
  inc si
  inc si
  div si
  test dl, cl
%else
  ; It is calculated by bitwise and-ing the result of rdtsc. (rdtsc returns the
  ; amount of CPU cycles since boot, which works okay as a cheap random number
  ; generator, and it's apparently supported on all x86 CPUs since the Pentium line)
  ;
  ; al = ! (rdtsc() & BombFreq)
cpu 686
  rdtsc
cpu 8086
  test al, BombFreq
%endif

  mov ax, VgaChar(Color.Veiled, Ascii.Bomb)
  
  ; If this cell isn't a bomb, then skip marking it as a bomb
  jnz .Iterated

  ; si = &TextBuf[y][x]
  call GetTextBufIndex

  ; Mark the current cell as a bomb
  ; Iterate over adjacent cells (directions)
  call AdjacentCells
.Iterated:

  ; Decrement x coordinate loop counter and continue if nonzero
  dec di
  jnz .LoopX

  ; Decrement y coordinate loop counter and continue if nonzero
  loop .LoopY
  ; cl and di are zeroed from the PopulateTextBuf loops above

;; Done populating the text buffer

  ; Set the initial cursor color for game loop. The dh register is now used to
  ; store the saved cell color that the cursor is on, since the cursor
  ; overwrites the cell color with the cursor color.
  xchg ax, dx

;; Main loop to process key presses and update state
GameLoop:

;; Detect win (a win occurs when every veiled cell is a mine)
DetectWin:
  mov bx, TextBuf.Size
  ; if (char != Ascii.Bomb && (color == Color.Veiled || color == Color.Flag)) {
  ;     break; // Didn't win yet :(
  ; }
  ; Use si register as cell pointer for win detection
%ifdef MDA_SUPPORT
  mov si, Mda.Buffer
.Loop:
  cs lodsw
%else
  xor si, si
.Loop:
  lodsw
%endif
  cmp al, Ascii.Bomb
  je .Continue
  cmp ah, Color.Unveiled
  jb .Break
.Continue:
  dec bx
  jne .Loop
  ; If loop completes without breaking, then we win! :)

;; Show game win screen
;;GameWin:
  mov ah, Color.GameWinText
  call GameEndHelper
  db 'GAME WIN'

.Break:
WaitRestart:
;; if bx == 0: Wait for restart key to be pressed, then restart game
  ; Get keystroke
  ; ah = BIOS scan code
  ; al = ASCII character
  ; http://www.delorie.com/djgpp/doc/rbinter/id/63/17.html
  mov ah, 0
  int 0x16
%ifdef DOS_QUIT
  dec ah
  je Quit
%endif
  dec bx
  js RunGame

  ; cl = y coord
  ; di = x coord

  ; si = cell pointer
  call GetTextBufIndex
  ; Apply saved cell color
%ifdef MDA_SUPPORT
  call SetColor
%else
  mov byte [si + 1], dh
%endif

  ; Didn't win yet
  xchg al, ah

;; Process key press. This is an if-else chain that runs code depending on the
;; key pressed.
CmpUp:
  ; Move cursor up
  dec cx
%ifdef DOS_QUIT
  sub al, Key.ScanCode.Up-1
%else
  sub al, Key.ScanCode.Up
%endif
  je WrapCursor
  inc cx
CmpDown:
  ; Move cursor down
  inc cx
  sub al, Key.ScanCode.Down - Key.ScanCode.Up
  je WrapCursor
  dec cx
CmpLeft:
  ; Move cursor left
  dec di
  sub al, Key.ScanCode.Left - Key.ScanCode.Down
  je WrapCursor
  inc di
CmpRight:
  ; Move cursor right
  inc di
  sub al, Key.ScanCode.Right - Key.ScanCode.Left
  je WrapCursor
  dec di
CmpEnter:
  sub al, Key.ScanCode.Enter - Key.ScanCode.Right
  jne CmpSpace
  ; Place flag by coloring current cell
%ifdef MDA_SUPPORT
 %ifdef UNDO_FLAG
  cmp byte [bp + si + Mda.Buffer - Mda.Screen.Offset + 1], Color.Unveiled
  jae GameLoop
  xor byte [bp + si + Mda.Buffer - Mda.Screen.Offset + 1], Color.Flag ^ Color.Veiled
  mov dx, [bp + si + Mda.Buffer - Mda.Screen.Offset]
 %else
  mov dh, Color.Flag
 %endif
  call SetColor
%else
 %ifdef UNDO_FLAG
  inc si
  cmp byte [si], Color.Unveiled
  jae GameLoop
  xor byte [si], Color.Flag ^ Color.Veiled
  mov dx, [si]
 %else
  mov dh, Color.Flag
  mov [si + 1], dh
 %endif
%endif
;  jmp GameLoop
CmpSpace:
  sub al, Key.ScanCode.Space - Key.ScanCode.Enter
  jne GameLoop

;; If the player pressed space, clear the current cell
ClearCell:
%ifdef EXPLOSION
 %ifdef MDA_SUPPORT
  mov al, [bp + si + Mda.Buffer - Mda.Screen.Offset]
 %else
  mov al, [si]
 %endif
  cmp al, Ascii.Bomb
%else
 %ifdef MDA_SUPPORT
  cmp byte [bp + si + Mda.Buffer - Mda.Screen.Offset], Ascii.Bomb
 %else
  cmp byte [si], Ascii.Bomb
 %endif
%endif
  je GameOver

  ; If cell is empty, run flood fill algorithm
  call Flood
%ifdef MDA_SUPPORT
GetCharAndColor:
  mov dx, [bp + si + Mda.Buffer - Mda.Screen.Offset]
%endif
jmpGameLoop:
  jmp GameLoop

;; Helper code for GameWin and GameOver; print a string in the center of the
;; text buffer, then wait for game to be restarted.
GameEndHelper:
  pop si
%ifdef MDA_SUPPORT
  mov di, TextBuf.Index(TextBuf.Height / 2, TextBuf.Width / 2 - (GameOverStr.End - GameOverStr) / 2) + Mda.Screen.Offset
%else
  mov di, TextBuf.Index(TextBuf.Height / 2, TextBuf.Width / 2 - (GameOverStr.End - GameOverStr) / 2)
%endif
  xor bx, bx
.String:
  cs lodsb
  cmp al, 'Z'
  ja WaitRestart
%ifdef MDA_SUPPORT
  mov [di + TextBuf.Index(TextBuf.Height / 2, TextBuf.Width / 2) - Mda.Screen.Offset], al
%endif
  stosw
  jmp .String

%ifdef DOS_QUIT
Quit:
  mov al,0x03           ; Restore text mode
  int 0x10
  int 0x20
  int 0x19
%endif

GameOver:
  ; If cell is bomb, game over :(
%ifdef EXPLOSION
  call UnveilCell
%endif  

;; Show game over screen

%ifdef UNVEIL_ON_GAME_OVER
 %ifdef MDA_SUPPORT
  call UnveilBomb
 %else
  mov cx, TextBuf.Size
  xor si, si
.Loop:
  ; Load VGA character into ax
  lodsw
  cmp al, Ascii.Bomb
  jne .Next
  and byte [si-1], Color.Unveiled
.Next:
  loop .Loop
 %endif
%endif
  mov ah, Color.GameOverText
  call GameEndHelper
GameOverStr:
  db 'GAME OVER'
GameOverStr.End:

;; Set y and x coordinates of cursor to zero if they are out of bounds
ResetCursor:
  xchg ax, cx
  xor di, di

WrapCursor:
  ; Get text buffer index (it changed)
  call GetTextBufIndex
  jae ResetCursor
  
  ; Draw cursor by changing cell to the cursor color, but save current color for
  ; restoring in the next iteration of the game loop.
%ifdef MDA_SUPPORT
  mov dx, Mda.Attr(Color.Cursor, Mda.Cursor)
  call SetScreenColor
  jmp GetCharAndColor
%else
  mov dh, Color.Cursor
  xchg byte [si + 1], dh
  jmp jmpGameLoop
%endif

%ifdef UNVEIL_ON_GAME_OVER
 %ifdef MDA_SUPPORT
FillScreen:
  mov cx, Mda.Veiled
UnveilBomb:
  mov si, Mda.Buffer
  mov bp, TextBuf.Height
  mov bx, 0 - TextBuf.Width + Mda.Buffer
.LoopLine:
  mov di, TextBuf.Width
  add bx, 2*TextBuf.Width
.Loop:
  mov ax, VgaChar(Color.Veiled, '0')
  or cl, cl
  js .Fill
  ; Load VGA character into ax
  mov ax,[cs:si]
  cmp al, Ascii.Bomb
  jne .Next
  and ah, Color.Unveiled
  mov cl, al
.Fill:
  mov [si], ax
  mov [cs:si], ax
  mov [bx + si + Mda.Screen.Offset - Mda.Buffer], cl
.Next:
  lodsw
  dec di
  jnz .Loop
  dec bp
  jnz .LoopLine
  ret
 %endif
%endif

;; Unveil a cell so it is visible on the screen
;;
;; Parameters:
;;   * si - cell pointer in text buffer
;;   * al - cell ASCII value
;; Returns:
;;   * dh - written VGA color code
UnveilCell:
  ; TLDR: Use xor magic to make the cells colored.
  ;
  ; We have three cases to consider:
  ;
  ; Case 1: the cell is a digit from 1-8
  ;
  ; The ASCII values '1', '2', '3', ..., '8' are 0x31, 0x32, 0x33, ..., 0x38. In
  ; other words, the upper nibble is always 0x3 and the lower nibble is the
  ; digit. We want the VGA color code to be `Color.Unveiled | digit`. For
  ; example, the color of a '3' cell would be 0xf3.
  ;
  ; We can accomplish this with the formula `cell_value ^ '0' ^ Color.Unveiled`.
  ; Xor-ing by '0' (0x30), clears the upper nibble of the cell value, leaving
  ; just the digit value. Xor-ing again by Color.Unveiled sets the upper nibble
  ; to 0xf, leading to the value `Color.Unveiled | digit`.
  ;
  ; Since xor is associative, this can be done in one operation, by xor-ing the
  ; cell value by ('0' ^ Color.Unveiled).
  ;
  ; Case 2: the cell is a bomb
  ;
  ; We don't really care about this case as long as the bomb is visible against
  ; the background. The bomb turns out to be bright yellow, oh well.
%ifdef MDA_SUPPORT
  mov dl, al
  mov [si], al
%endif
  xor al, '0' ^ Color.Unveiled
  mov dh, al
%ifdef MDA_SUPPORT
SetColor:
  mov [bp + si + Mda.Buffer - Mda.Screen.Offset + 1], dh
  cmp dh, Color.Flag
  ja .NotVeiled
  mov dl, Mda.Veiled
  jne .NotVeiled
  mov dl, Mda.Flag
.NotVeiled:
SetScreenColor:
  mov [bx + si], dl
%endif
  mov [si + 1], dh
  ret

;; Flood fill empty cells
;;
;; Parameters:
;;   * cl - cell y coordinate
;;   * di - cell x coordinate
;;   * si - cell pointer in text buffer
FloodStep:
%ifdef MDA_SUPPORT
  mov al, [bp + si + Mda.Buffer - Mda.Screen.Offset]
%else
  mov al, [si]
%endif
  cmp al, '0'

  ; Base case: nonempty cell unveiled and stop recursion
  ja UnveilCell

  mov ax, VgaChar(Color.Unveiled, ' ')
  
  ; Base case: we visited this cell already or bomb
  jb Ret

  ; Body: mark cell as visited and empty
%ifdef MDA_SUPPORT
  mov [si], ax
  mov [bx + si], al
AdjacentCells:
  mov [bp + si + Mda.Buffer - Mda.Screen.Offset], ax
%else
AdjacentCells:
  mov [si], ax
%endif

  ; Recursive case: flood adjacent cells

  ; Flood left-row
  dec di
  call UpAndFloodRow

  ; Flood center-row
  call FloodNextRow

  ; Flood right-row
  call FloodNextRow
  
  ; re-center
  dec cx
  dec di

;; Compute the text buffer index from y and x coordinates
;;
;; si = &TextBuf[cl = y][di = x]
;;
;; This computes the equivalent of the TextBuf.Index(y, x) macro, but at runtime
;;
;; Parameters:
;;   * cl - y coordinate
;;   * di - x coordinate
;; Returns:
;;   * si - text buffer index
;;   * carry - x and y are valid coordinates
GetTextBufIndex:
  xchg ax, si
  mov al, TextBuf.Width * 2
  imul cl
  xchg ax, si
%ifdef MDA_SUPPORT
  lea bx, [si + TextBuf.Width - Mda.Screen.Offset]
  lea si, [bx + di - TextBuf.Width]
  add si, di
 %ifdef LAZY_CHECK
  cmp si, TextBuf.Size * 2 + Mda.Screen.Offset
 %endif
%else
  add si, di
  add si, di
 %ifdef LAZY_CHECK
  cmp si, TextBuf.Size * 2
 %endif
%endif
%ifndef LAZY_CHECK
  ; Base case: bounds check y
  cmp cl, TextBuf.Height
  jae Ret
  ; Base case: bounds check x
  cmp di, TextBuf.Width
%endif
Ret:
  ret

FloodNextRow:
  inc di
  dec cx
UpAndFloodRow:
  dec cx
FloodRow:
  call Flood
  call FloodDown
FloodDown:
  inc cx
Flood:
  ; si = &TextBuf[y][x]
  call GetTextBufIndex
  jae Ret
  
  cmp al, Ascii.Bomb
  jne FloodStep
Counter:
  ; If adjacent cell is a bomb, skip digit incrementing
%ifdef MDA_SUPPORT
  cmp [bp + si + Mda.Buffer - Mda.Screen.Offset], al
%else
  cmp [si], al
%endif
  je .IsMine
  ; The adjacent cell is a 0-7 digit and not a bomb. Add 1 to the cell.
  ; This gradually accumulates to the
  ; amount of neighboring bombs and represents the number cells in the
  ; minesweeper game.
%ifdef MDA_SUPPORT
  inc byte [bp + si + Mda.Buffer - Mda.Screen.Offset]
%else
  inc byte [si]
%endif
.IsMine:
  ret

;; Print program size at build time
%assign CodeSize $ - $$
%assign Size $ - BootMine
%warning Code is Size bytes

CodeEnd:
  ; Pad to size of boot sector, minus the size of a word for the boot sector
  ; magic value. If the code is too big to fit in a boot sector, the `times`
  ; directive uses a negative value, causing a build error.
  times (BootSector.Size - WordSize) - CodeSize db 0

  ; Boot sector magic
  dw 0xaa55
